/*    puttyappui.cpp
 *
 * Putty UI Application UI class
 *
 * Copyright 2002 Petteri Kangaslampi
 *
 * See license.txt for full copyright and license information.
*/

#include <eikenv.h>
#include <eikon.hrh>
#include <ckninfo.h>
#include <cknconf.h>
#include <es_sock.h>
#include <eikmenup.h>
#include <eikdoc.h>

#include "puttyappui.h"
#include "puttyappview.h"
#include "puttyengine.h"
#include "puttyui.hrh"
#include "connectiondialog.h"
#include <puttyui.rsg>
extern "C" {
#include "putty.h"
}

const TInt KInitAudioLength = 32000;

_LIT(KAssertPanic, "puttyappui.cpp");
#define assert(x) __ASSERT_ALWAYS(x, User::Panic(KAssertPanic, __LINE__))

_LIT(KRandomFile,"random.dat");

CPuttyAppUi::CPuttyAppUi() : iAudioRecordDes(NULL, 0, 0) {
    iTermWidth = 80;
    iTermHeight = 24;
    iLargeFont = EFalse;
}


void CPuttyAppUi::ConstructL() {

    CEikonEnv::Static()->DisableExitChecks(ETrue);
    BaseConstructL();

    iFatalErrorPanic =
        CEikonEnv::Static()->AllocReadResourceL(R_STR_FATAL_ERROR);

    iDialler = CDialler::NewL(this);
    iRecorder = CAudioRecorder::NewL(this);
    
    iAppView = new (ELeave) CPuttyAppView(this);
    iAppView->ConstructL(ClientRect());
    AddToStackL(iAppView);

    iAppView->Terminal()->SetGrayed(ETrue);
}


CPuttyAppUi::~CPuttyAppUi() {
    delete iAppView;
    delete iRecorder;
    delete iAudio;
    delete iDialler;
    if ( iEngine ) {
        iEngine->Close();
    }
    delete iFatalErrorPanic;
}


TBool CPuttyAppUi::ProcessCommandParametersL(TApaCommand /*aCommand*/,
                                             TFileName & /*aDocumentName*/,
                                             const TDesC8 &aTail) {
    // Find the engine object
    TLex8 lexx(aTail);
    TUint eng;
    User::LeaveIfError(lexx.Val(eng, EHex));
    iEngine = (MPuttyEngine*) eng;

    // Determine application installation path
    TFileName appDll = iDocument->Application()->AppFullName();
    TParse parsa;
    User::LeaveIfError(parsa.SetNoWild(appDll, NULL, NULL));
    TFileName dataPath = parsa.DriveAndPath();

    // If the application is in ROM (drive Z:), put settings on C:
    if ( (dataPath[0] == 'Z') || (dataPath[0] == 'z') ) {
        dataPath[0] = 'C';
    }

    // Make sure the path ends with a backslash
    if ( dataPath[dataPath.Length()-1] != '\\' ) {
        dataPath.Append('\\');
    }

    // Check if the random seed already exists. If not, we are probably running
    // PuTTY for the first time and the generator should be seeded
    TFileName seedFile = dataPath;    
    seedFile.Append(KRandomFile);
    RFile f;
    TBool randomExists = EFalse;
    TInt error = f.Open(CEikonEnv::Static()->FsSession(), seedFile,
                        EFileRead | EFileShareReadersOnly);
    if ( error == KErrNone ) {
        randomExists = ETrue;
        f.Close();
    }
    
    // Initialize the engine
    iEngine->Init(this, dataPath);
    iEngine->SetTerminalSize(iTermWidth, iTermHeight);

    // Initialize the random number generator if necessary
    if ( !randomExists ) {
        
        if ( CCknConfirmationDialog::RunDlgLD(R_STR_INITIAL_RANDOM_TITLE,
                                              R_STR_INITIAL_RANDOM_TEXT,
                                              NULL,
                                              R_STR_OK_CONFIRM) ) {
            HandleCommandL(ECmdInitRandomGenerator);
        }
    }

    return EFalse;
}


void CPuttyAppUi::HandleCommandL(TInt aCommand) {

    switch (aCommand) {

        case ECmdConnect: {
            if ( iState != EStateNone ) {
                break;
            }
            assert(iEngine);
            Config *cfg = iEngine->GetConfig();

            THostName hostName;
            CConnectionDialog *dlg = new (ELeave) CConnectionDialog(hostName);
            
            if ( dlg->ExecuteLD(R_CONNECTION_DIALOG)) {
                char *c = cfg->host;
                for ( TInt i = 0; i < hostName.Length(); i++ ) {
                    *c++ = (char) hostName[i];
                }
                *c++ = 0;

                TRAPD(error, CEikonEnv::Static()->BusyMsgL(R_STR_DIALING));
                iState = EStateDialing;
                iDialler->DialL();
            }
            break;
        }

        case ECmdLargeFont: {
            if ( iLargeFont ) {
                iLargeFont = EFalse;
            } else {
                iLargeFont = ETrue;
            }
            iAppView->SetFontL(iLargeFont);
            break;
        }
        
        case ECmdFullScreen: {
            if ( iFullScreen ) {
                iFullScreen = EFalse;
            } else {
                iFullScreen = ETrue;
            }
            iAppView->SetFullScreenL(iFullScreen);
            break;
        }

        case ECmdInitRandomGenerator: {
            if ( iRecording ) {
                break;
            }
            
            if ( CCknConfirmationDialog::RunDlgLD(R_STR_RECORD_CONFIRM_TITLE,
                                                  R_STR_RECORD_CONFIRM_TEXT,
                                                  NULL,
                                                  R_STR_OK_CONFIRM) ) {
                delete iAudio;
                iAudio = NULL;
                iAudio = HBufC8::NewL(KInitAudioLength);
                iAudioRecordDes.Set(iAudio->Des());
                iRecorder->RecordL(iAudioRecordDes);
                iRecording = ETrue;
                CEikonEnv::Static()->BusyMsgL(R_STR_RECORDING);
            }
            
            break;
        }
        
	case EEikCmdExit: 
            Exit();
            break;
    }    
}


void CPuttyAppUi::DynInitMenuPaneL(TInt aResourceId, CEikMenuPane *aMenuPane) {
    
    if ( aResourceId == R_PUTTY_VIEW_MENU ) {
        if ( iLargeFont ) {
            aMenuPane->SetItemButtonState(ECmdLargeFont, EEikMenuItemSymbolOn);
        } else {
            aMenuPane->SetItemButtonState(ECmdLargeFont,
                                          EEikMenuItemSymbolIndeterminate);
        }
        
        if ( iFullScreen ) {
            aMenuPane->SetItemButtonState(ECmdFullScreen,
                                          EEikMenuItemSymbolOn);
        } else {
            aMenuPane->SetItemButtonState(ECmdFullScreen,
                                          EEikMenuItemSymbolIndeterminate);
        }
    }
}


// MDialObserver::DialCompleted()
void CPuttyAppUi::DialCompleted(TInt aError) {

    assert(iState == EStateDialing);
    CEikonEnv *eenv = CEikonEnv::Static();    
    eenv->BusyMsgCancel();
    
    if ( aError != KErrNone ) {
        HBufC *msg = HBufC::NewLC(512);
        TPtr msgp = msg->Des();
        HBufC *err = HBufC::NewLC(128);
        TPtr errp = err->Des();
        
        eenv->GetErrorText(errp, aError);
        msgp.Format(*eenv->AllocReadResourceLC(R_STR_DIALUP_FAILED),
                    &errp);
        HBufC *title = eenv->AllocReadResourceLC(R_STR_CONNECTION_ERROR_TITLE);
        
        CCknInfoDialog::RunDlgLD(*title, msgp);
        CleanupStack::PopAndDestroy(4); // title, formatstring, err, msg

        iState = EStateNone;
        return;
    }

    TRAPD(error, eenv->BusyMsgL(R_STR_CONNECTING));
    if ( error != KErrNone ) {
        User::Panic(*iFatalErrorPanic, error);
    }
    iState = EStateConnecting;

    TInt err = iEngine->Connect();
    if ( err != KErrNone ) {
        TBuf<128> msg;
        iEngine->GetErrorMessage(msg);
        ConnectionErrorL(msg);
        Exit();
    }

    iState = EStateConnected;
    eenv->BusyMsgCancel();
    iAppView->Terminal()->SetGrayed(EFalse);
}


// MRecordObserver::RecordCompleted()
void CPuttyAppUi::RecordCompleted(TInt aError) {

    CEikonEnv *eenv = CEikonEnv::Static();
    eenv->BusyMsgCancel();

    // If the audio device was reserved, prompt to try again
    if ( aError == KErrInUse ) {

        if ( CCknConfirmationDialog::RunDlgLD(R_STR_RECORD_ERROR_TITLE,
                                              R_STR_AUDIO_DEVICE_IN_USE,
                                              NULL,
                                              R_STR_OK_CONFIRM) ) {
            iRecorder->CancelRecord();
            iAudioRecordDes.SetLength(0);
            iRecorder->RecordL(iAudioRecordDes);
            iRecording = ETrue;
            CEikonEnv::Static()->BusyMsgL(R_STR_RECORDING);
            return;
        }
        
    } else if ( aError != KErrNone ) {
        
        // Handle other errors                    
        iRecorder->CancelRecord();
        
        HBufC *title =
            CEikonEnv::Static()->AllocReadResourceLC(R_STR_RECORD_ERROR_TITLE);

        HBufC *msg = HBufC::NewLC(512);
        TPtr msgp = msg->Des();
        HBufC *err = HBufC::NewLC(128);
        TPtr errp = err->Des();        
        eenv->GetErrorText(errp, aError);
        msgp.Format(*eenv->AllocReadResourceLC(R_STR_RECORD_FAILED),
                    &errp);
        
        CCknInfoDialog::RunDlgLD(*title, msgp);
        CleanupStack::PopAndDestroy(4); // formatstring, err, msg, title

    } else {
        // Use the data as random number seed noise.
        iEngine->AddRandomNoise(iAudioRecordDes);
        eenv->InfoMsg(R_STR_RANDOMIZED);
    }

    delete iAudio;
    iAudio = NULL;
    iRecording = EFalse;        
}


// MPuttyClient::DrawText()
void CPuttyAppUi::DrawText(TInt aX, TInt aY, const TDesC &aText,
                           TBool aBold, TBool aUnderline,
                           TRgb aForeground, TRgb aBackground) {
    
    iAppView->Terminal()->DrawText(aX, aY, aText, aBold, aUnderline,
                                   aForeground, aBackground);
}


// MPuttyClient::ConnectionError()
void CPuttyAppUi::ConnectionError(const TDesC &aMessage) {

    TRAPD(error, ConnectionErrorL(aMessage));
    if ( error != KErrNone ) {
        User::Panic(*iFatalErrorPanic, error);
    }
    iAppView->Terminal()->SetGrayed(ETrue);
}

void CPuttyAppUi::ConnectionErrorL(const TDesC &aMessage) {

    HBufC *title =
        CEikonEnv::Static()->AllocReadResourceLC(R_STR_CONNECTION_ERROR_TITLE);
    CCknInfoDialog::RunDlgLD(*title, aMessage);
    CleanupStack::PopAndDestroy();
    Exit();
}


// MPuttyClient::FatalError()
void CPuttyAppUi::FatalError(const TDesC &aMessage) {

    TRAPD(error, FatalErrorL(aMessage));
    if ( error != KErrNone ) {
        User::Panic(*iFatalErrorPanic, error);
    }
}

void CPuttyAppUi::FatalErrorL(const TDesC &aMessage) {

    HBufC *title =
        CEikonEnv::Static()->AllocReadResourceLC(R_STR_FATAL_ERROR_TITLE);
    CCknInfoDialog::RunDlgLD(*title, aMessage);
    CleanupStack::PopAndDestroy();
    Exit();
}


// MPuttyClient::ConnectionClosed()
void CPuttyAppUi::ConnectionClosed() {
    CEikonEnv::Static()->InfoMsg(R_STR_CONNECTION_CLOSED);
    iAppView->Terminal()->SetGrayed(ETrue);
}


// MPuttyClient::UnknownHostKey()
//THostKeyResponse CPuttyAppUi::UnknownHostKey(const TDesC &aFingerprint) {
MPuttyClient::THostKeyResponse CPuttyAppUi::UnknownHostKey(
    const TDesC &aFingerprint) {
    
    MPuttyClient::THostKeyResponse resp = EAbadonConnection;
    TRAPD(error, resp = HostKeyDialogL(aFingerprint,
                                       R_STR_UNKNOWN_HOST_KEY_TITLE,
                                       R_STR_UNKNOWN_HOST_KEY_DLG_FMT));
    if ( error != KErrNone ) {
        User::Panic(*iFatalErrorPanic, error);
    }
    return resp;
}


// MPuttyClient::DifferentHostKey()
MPuttyClient::THostKeyResponse CPuttyAppUi::DifferentHostKey(
    const TDesC &aFingerprint) {
    
    MPuttyClient::THostKeyResponse resp = EAbadonConnection;
    TRAPD(error, resp = HostKeyDialogL(aFingerprint,
                                       R_STR_DIFFERENT_HOST_KEY_TITLE,
                                       R_STR_DIFFERENT_HOST_KEY_DLG_FMT));
    if ( error != KErrNone ) {
        User::Panic(*iFatalErrorPanic, error);
    }
    return resp;
}


MPuttyClient::THostKeyResponse CPuttyAppUi::HostKeyDialogL(
    const TDesC &aFingerprint, TInt aDialogTitleRes, TInt aDialogFormatRes) {
    CEikonEnv *env = CEikonEnv::Static();

    HBufC *title = env->AllocReadResourceLC(aDialogTitleRes);
    HBufC *fmt = env->AllocReadResourceLC(aDialogFormatRes);

    HBufC *contents = HBufC::NewLC(fmt->Length() + aFingerprint.Length());
    contents->Des().Format(*fmt, &aFingerprint);

    TInt res = CCknConfirmationDialog::RunDlgLD(
        *title, *contents, R_UNKNOWN_HOST_KEY_DLG_BUTTONS);

    MPuttyClient::THostKeyResponse resp = EAbadonConnection;

    switch ( res ) {
        case ECmdHostKeyAcceptAndSave:
            resp = EAcceptAndStore;
            break;
        
        case ECmdHostKeyAcceptOnce:
            resp = EAcceptTemporarily;
            break;

        case ECmdHostKeyReject:
            resp = EAbadonConnection;
            break;

        default:
            assert(EFalse);
    }

    CleanupStack::PopAndDestroy(3); // contents, fmt, title

    return resp;
}


// MPuttyClient::AcceptCipher()
TBool CPuttyAppUi::AcceptCipher(const TDesC &aCipherName,
                                TCipherDirection aDirection) {
    TBool resp = EFalse;
    TRAPD(error, resp = AcceptCipherL(aCipherName,
                                      aDirection));
    if ( error != KErrNone ) {
        User::Panic(*iFatalErrorPanic, error);
    }
    return resp;
}

TBool CPuttyAppUi::AcceptCipherL(const TDesC &aCipherName,
                                 TCipherDirection aDirection) {
    
    CEikonEnv *env = CEikonEnv::Static();

    HBufC *title = env->AllocReadResourceLC(R_STR_ACCEPT_CIPHER_TITLE);
    HBufC *fmt = env->AllocReadResourceLC(R_STR_ACCEPT_CIPHER_DLG_FMT);
    HBufC *dir = NULL;

    switch ( aDirection ) {
        case EBothDirections:
            dir = env->AllocReadResourceLC(R_STR_ACCEPT_CIPHER_DIR_BOTH);
            break;

        case EClientToServer:
            dir = env->AllocReadResourceLC(
                R_STR_ACCEPT_CIPHER_CLIENT_TO_SERVER);
            break;

        case EServerToClient:
            dir = env->AllocReadResourceLC(
                R_STR_ACCEPT_CIPHER_SERVER_TO_CLIENT);
            break;

        default:
            assert(EFalse);
    }

    HBufC *contents = HBufC::NewLC(fmt->Length() + aCipherName.Length() +
                                   dir->Length());
    contents->Des().Format(*fmt, &aCipherName, dir);

    TBool res = CCknConfirmationDialog::RunDlgLD(*title, *contents);

    CleanupStack::PopAndDestroy(4); // contents, fir, fmt, title

    return res;
}


// MTerminalObserver::TerminalSizeChanged()
void CPuttyAppUi::TerminalSizeChanged(TInt aWidth, TInt aHeight) {
    assert((aWidth > 1) && (aHeight > 1));
    iTermWidth = aWidth;
    iTermHeight = aHeight;
    if ( iEngine ) {
        iEngine->SetTerminalSize(aWidth, aHeight);
    }
}


// MTerminalObserver::KeyPressed()
void CPuttyAppUi::KeyPressed(TKeyCode aCode, TUint aModifiers) {
    if ( iEngine ) {
        iEngine->SendKeypress(aCode, aModifiers);
    }
}


// MTerminalObserver::RePaintWindow()
void CPuttyAppUi::RePaintWindow() {
    if ( iEngine ) {
        iEngine->RePaintWindow();
    }
}
