/*    terminalcontrol.cpp
 *
 * A terminal UI control class
 *
 * Copyright 2002 Petteri Kangaslampi
 *
 * See license.txt for full copyright and license information.
*/

#include <e32std.h>
#include <gdi.h>
#include <eikenv.h>
#include "terminalcontrol.h"

_LIT(KTerminalFont, "Terminal");
_LIT(KTerminalControl, "terminalcontrol.cpp");
static const TInt KSmallFontHeight = 8;
static const TInt KLargeFontHeight = 14;
static const TInt KDefaultLargeFont = EFalse;

#define assert(x) __ASSERT_ALWAYS(x, User::Panic(KTerminalControl, __LINE__))

// Small: 6x8
// Large: 7x14
// The bold version is one pixel wider, and thus unusable!


CTerminalControl::CTerminalControl(MTerminalObserver *anObserver) {
    iObserver = anObserver;
    iInDraw = EFalse;
    iGrayed = EFalse;
}


CTerminalControl::~CTerminalControl() {

    if ( iFont != NULL ) {
        CEikonEnv::Static()->ScreenDevice()->ReleaseFont(iFont);
        iFont = NULL;
    }
}


void CTerminalControl::ConstructL(const TRect &aRect,
                                  RWindow &aContainerWindow) {

    SetContainerWindowL(aContainerWindow);
    
    // Set the default font
    SetFontL(KDefaultLargeFont);

    SetRect(aRect);
    Resize();
}


void CTerminalControl::SetFontL(TBool aLargeFont) {

    TInt fontHeight = KSmallFontHeight;    
    if ( aLargeFont ) {
        fontHeight = KLargeFontHeight;
    }

    // Release old fonts
    if ( iFont != NULL ) {
        CEikonEnv::Static()->ScreenDevice()->ReleaseFont(iFont);
        iFont = NULL;
    }
    
    // Get the font
    CGraphicsDevice *dev = CEikonEnv::Static()->ScreenDevice();    
    TInt twips = dev->VerticalPixelsToTwips(fontHeight);
    TFontSpec spec(KTerminalFont, twips);
    User::LeaveIfError(dev->GetNearestFontInTwips(iFont, spec));

    // Get font info and resize terminal
    iFontHeight = iFont->HeightInPixels();
    iFontWidth = iFont->WidthZeroInPixels();
}


void CTerminalControl::SetGrayed(TBool aGrayed) {
    iGrayed = aGrayed;
    DrawDeferred();
}


// Resizes the terminal in response to size or font change
void CTerminalControl::Resize() {

    TRect rect = Rect();
    TInt width = rect.iBr.iX - rect.iTl.iX;
    TInt height = rect.iBr.iY - rect.iTl.iY;

    assert((iFontWidth > 0) && (iFontHeight > 0));
    assert((width >= iFontWidth) && (height >= iFontHeight));
    TInt charWidth = width / iFontWidth;
    TInt charHeight = height / iFontHeight;
    iObserver->TerminalSizeChanged(charWidth, charHeight);
}


// Draws text on the terminal window. The coordinates are zero-based
// character coordinates inside the terminal.
void CTerminalControl::DrawText(TInt aX, TInt aY, const TDesC &aText,
                                TBool /*aBold*/, TBool aUnderline,
                                TRgb aForeground, TRgb aBackground) {

    // Don't draw text while grayed out
    if ( iGrayed ) {
        return;
    }

    // We'll draw the text inside a filled box. This should avoid flickering
    // compared to filling first and drawing afterwards

    CWindowGc &gc = SystemGc();
    if ( !iInDraw ) {
        gc.Activate(Window());
    }
    gc.UseFont(iFont);

    // Brush is used for background
    gc.SetBrushColor(aBackground);
    gc.SetBrushStyle(CGraphicsContext::ESolidBrush);

    // Pen is used for foreground
    gc.SetPenColor(aForeground);
    gc.SetPenStyle(CGraphicsContext::ESolidPen);

    // Handle underline. Sadly we cannot use a bold version of the Terminal
    // font to do real bold, since it is one pixel wider than the normal
    // version
    if ( aUnderline ) {
        gc.SetUnderlineStyle(EUnderlineOn);
    }    

    // Draw the text inside a box
    TInt ascent = iFont->AscentInPixels();
    TInt x = Rect().iTl.iX;
    TInt y = Rect().iTl.iY;
    TRect textRect(x + aX * iFontWidth, y + aY * iFontHeight,
                   x + (aX + aText.Length()) * iFontWidth,
                   y + (aY+1) * iFontHeight);
    gc.DrawText(aText, textRect, ascent);

    // Reset underline attribute
    if ( aUnderline ) {
        gc.SetUnderlineStyle(EUnderlineOff);
    }

    gc.DiscardFont();
    if ( !iInDraw ) {
        gc.Deactivate();
    }
}


// CCoeControl::OfferKeyEventL()
TKeyResponse CTerminalControl::OfferKeyEventL(const TKeyEvent &aKeyEvent,
                                              TEventCode aType) {

    if ( aType == EEventKey ) {
        TUint code = aKeyEvent.iCode;

        // Do not eat Menu or Chr
        if ( (code == EKeyMenu) || (code == EKeyLeftFunc) ||
             (code == EKeyRightFunc) ) {
            return EKeyWasNotConsumed;
        }

        // Pass keypress to the observer (and thus to the PuTTY engine)
        iObserver->KeyPressed((TKeyCode)code, aKeyEvent.iModifiers);
        return EKeyWasConsumed;
    }

    return EKeyWasNotConsumed;
}


// CCoeControl::Draw()
void CTerminalControl::Draw(const TRect & /*aRect*/) const {

    ((CTerminalControl*)this)->iInDraw = ETrue;
    CWindowGc &gc = SystemGc();
    gc.Reset();

    if ( iGrayed ) {
        gc.SetBrushStyle(CGraphicsContext::ESolidBrush);
        gc.SetBrushColor(TRgb(0xcccccc));
        gc.SetPenStyle(CGraphicsContext::ENullPen);
        gc.DrawRect(Rect());
        iObserver->RePaintWindow();
    }
    else {           
        gc.SetBrushStyle(CGraphicsContext::ESolidBrush);
        gc.SetBrushColor(KRgbWhite);
        gc.SetPenStyle(CGraphicsContext::ENullPen);
        gc.DrawRect(Rect());
        iObserver->RePaintWindow();
    }
    ((CTerminalControl*)this)->iInDraw = EFalse;
}


// CCoeControl::SizeChanged()
void CTerminalControl::SizeChanged() {
    Resize();
}


// CCoeControl::InputCapabilities()
TCoeInputCapabilities CTerminalControl::InputCapabilities() const {
    return TCoeInputCapabilities(TCoeInputCapabilities::EAllText |
                                 TCoeInputCapabilities::ENavigation);
}
