/*    terminalcontrol.h
 *
 * A terminal UI control class
 *
 * Copyright 2002 Petteri Kangaslampi
 *
 * See license.txt for full copyright and license information.
*/

#ifndef __TERMINALCONTROL_H__
#define __TERMINALCONTROL_H__

#include <e32std.h>
#include <coecntrl.h>


class MTerminalObserver {

public:
    // Terminal size has been changed
    virtual void TerminalSizeChanged(TInt aWidth, TInt aHeight) = 0;

    // A key has been pressed
    virtual void KeyPressed(TKeyCode aCode, TUint aModifiers) = 0;

    // Force window repaint. The user must draw each line again with DrawText()
    virtual void RePaintWindow() = 0;
};


class CFont;


/**
 * A terminal UI control class. Manages a terminal control on the screen,
 * taking care of drawing the text and delivering keypresses to the user.
 * The terminal user must implement the MTerminalObserver interface.
 */
class CTerminalControl : public CCoeControl {

public:    
    /** 
     * Terminal control constructor.
     * 
     * @param anObserver The observer class to use.
     */
    CTerminalControl(MTerminalObserver *anObserver);
    
    /** 
     * Second-phase constructor. Must be called before the control is used.
     * FIXME: Replace with NewL.
     * 
     * @param aRect Initial terminal control rectangle on screen. Use SetRect()
     *              to change the terminal size.
     * @param aContainerWindow The window that contains this control.
     */
    void ConstructL(const TRect &aRect, RWindow &aContainerWindow);

    /** 
     * Destructor.
     */
    ~CTerminalControl();

    /** 
     * Draws text on the terminal window. The coordinates are zero-based
     * character coordinates inside the terminal. There is no clipping --
     * the terminal control assumes that the user takes care of this.
     *
     * Cursors can be drawn as text with a different background color.
     * 
     * @param aX Text X-coordinate
     * @param aY Text Y-coordinate
     * @param aText The text to draw
     * @param aBold ETrue if text should be bold (not supported currently)
     * @param aUnderline ETrue if text should be underlined.
     * @param aForeground Text foreground color
     * @param aBackground Text background color. The text is drawn inside a
     *                    solid rectangle colored with this color.
     */
    virtual void DrawText(TInt aX, TInt aY, const TDesC &aText, TBool aBold,
                          TBool aUnderline, TRgb aForeground,
                          TRgb aBackground);

    /** 
     * Changes the terminal font.
     * 
     * @param aLargeFont ETrue to use the large terminal font
     */
    void SetFontL(TBool aLargeFont);


    /** 
     * Sets the terminal control to be grayed out or normal. The terminal
     * control can be grayed out when there is no connection, and it cannot
     * be redrawn properly.
     * 
     * @param aGrayed True if the terminal should be grayed out.
     */
    void SetGrayed(TBool aGrayed);


    /** \fn void SetRect(const TRect aRect);
     *
     * Sets the terminal control size and position on screen. The actual size
     * used is rounded down from the argument to the nearest size in full
     * characters.
     *
     * @param aRect New terminal window size and position.
     */

    // CCoeControl methods
    virtual TKeyResponse OfferKeyEventL(const TKeyEvent &aKeyEvent,
                                        TEventCode aType);
    virtual void Draw(const TRect &aRect) const;
    virtual void SizeChanged();
    virtual TCoeInputCapabilities InputCapabilities() const;

private:
    void Resize();
    
    CFont *iFont;
    CFont *iBoldFont;
    TInt iFontWidth, iFontHeight; // font char dimensions in pixels
    MTerminalObserver *iObserver;
    TBool iInDraw;
    TBool iGrayed;
};


#endif
