/*    puttyengineimp.cpp
 *
 * PuTTY engine implementation class
 *
 * Copyright 2002 Petteri Kangaslampi
 *
 * See license.txt for full copyright and license information.
*/

#include <e32std.h>
#include <assert.h>
#include <stdlib.h>
#include "charutil.h"
#include "puttyengineimp.h"
#include "epocmemory.h"
#include "epocnoise.h"
#include "epocnet.h"
#include "epocstore.h"
#include "terminalkeys.h"
extern "C" {
#include "storage.h"
#include "ssh.h"
}


_LIT(KOutOfMemory, "Out of memory");
static CPuttyEngineImp *engine;

static const unsigned char KDefaultColors[22][3] = {
    { 0,0,0 }, { 128,128,128 }, { 255,255,255 }, { 255,255,255 }, { 0,0,0 },
    { 128,128,192 }, { 0,0,0 }, { 85,85,85 }, { 187,0,0 }, { 255,85,85 },
    { 0,187,0 }, { 85,255,85 }, { 187,187,0 }, { 255,255,85 }, { 0,0,187 },
    { 85,85,255 }, { 187,0,187 }, { 255,85,255 }, { 0,187,187 },
    { 85,255,255 }, { 187,187,187 }, { 192,192,192 }
};

struct backend_list backends[] = {
    {Config::PROT_SSH, "ssh", &ssh_backend},
    {0, NULL}
};


// Factory methods

CPuttyEngineImp *CPuttyEngineImp::NewL() {
    CPuttyEngineImp *self = new (ELeave) CPuttyEngineImp;
    CleanupStack::PushL(self);
    self->ConstructL();
    CleanupStack::Pop();
    return self;
}

CPuttyEngineImp *CPuttyEngineImp::NewLC() {
    CPuttyEngineImp *self = new (ELeave) CPuttyEngineImp;
    CleanupStack::PushL(self);
    self->ConstructL();
    return self;
}


// Two-phase construction
CPuttyEngineImp::CPuttyEngineImp() : CActive(EPriorityNormal) {
    iState = EStateNone;
    iTermWidth = 80;
    iTermHeight = 24;
}

void CPuttyEngineImp::ConstructL() {
}


// Destruction
CPuttyEngineImp::~CPuttyEngineImp() {
    if ( iState != EStateNone ) {
        Close();
    }
    delete iTermUpdatePeriodic;
}


// MPuttyEngine::Init()
void CPuttyEngineImp::Init(MPuttyClient *aClient, const TDesC &aDataPath) {
    assert(iState == EStateNone);
    iClient = aClient;
    iConnError = NULL;
    iFatalConnError = NULL;
    iNumSockets = 0;
    
    // Initialize Symbian-port bits
    epoc_memory_init();
    epoc_store_init(aDataPath);
    epoc_noise_init();
    random_init();

    // Set defaults
    has_focus = 1;
    flags = FLAG_INTERACTIVE;
    default_protocol = Config::PROT_SSH;
    default_port = 22;    
    do_defaults(NULL, &cfg);
    default_protocol = cfg.protocol;
    default_port = cfg.port;
    strcpy(cfg.line_codepage, "ISO-8859-15");
    cfg.vtmode = VT_UNICODE;
    cfg.sshprot = 1;
    for ( TInt i = 0; i < 22; i++ ) {
        cfg.colours[i][0] = KDefaultColors[i][0];
        cfg.colours[i][1] = KDefaultColors[i][1];
        cfg.colours[i][2] = KDefaultColors[i][2];
    }

    engine = this;
    iState = EStateInitialized;
    
    CActiveScheduler::Add(this);
}


// MPuttyEngine::Close()
void CPuttyEngineImp::Close() {
    if ( (iState == EStateConnected) || (iState == EStateDisconnected) ) {
        Disconnect();
    } else {
        assert(iState == EStateInitialized);
    }

    // Save random number generator seed. The Symbian OS implementation won't
    // actually save it unless the generator has been seeded properly, so this
    // can be called unconditionally.
    random_save_seed();

    // Uninitialize Symbian stuff
    epoc_noise_free();
    epoc_store_free();
    epoc_memory_free();

    CloseSTDLIB();

    iState = EStateNone;
    
    Cancel();
}


// MPuttyEngine::GetConfig()
Config *CPuttyEngineImp::GetConfig() {
    return &cfg;
}


// MPuttyEngine::Connect()
TInt CPuttyEngineImp::Connect() {

    assert(iState == EStateInitialized);

    // Convert palette from the config to our internal palette
    static const TInt palettemap[24] = {
	6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17,
        18, 19, 20, 21, 0, 1, 2, 3, 4, 4, 5, 5
    };
    for ( TInt c = 0; c < 24; c++ ) {
        TInt idx = palettemap[c];
        iDefaultPalette[c].SetRed(cfg.colours[idx][0]);
        iDefaultPalette[c].SetGreen(cfg.colours[idx][1]);
        iDefaultPalette[c].SetBlue(cfg.colours[idx][2]);
    }
    putty_palette_reset();
    
    // Initialize charset conversion tables
    init_ucs_tables();

    // Select the protocol to use
    back = NULL;
    for ( TInt i = 0; backends[i].backend != NULL; i++ ) {
        if ( backends[i].protocol == cfg.protocol ) {
            back = backends[i].backend;
            break;
        }
    }
    if ( back == NULL ) {
        fatalbox("Unsupported backend");
    }

    // Initialize networking
    sk_init();
    iNumSockets = 0;
    sk_set_watcher(this);

    // Connect
    char *realhost;
    iConnError = back->init(cfg.host, cfg.port, &realhost, cfg.tcp_nodelay);
    if ( iConnError ) {
        sk_set_watcher(NULL);
        sk_cleanup();
        return KErrGeneral;
    }
    sfree(realhost);

    // Initialize terminal
    term_init();
    term_size(iTermHeight, iTermWidth, cfg.savelines);

    // Start a periodic timer to update the terminal (FIXME?)
    delete iTermUpdatePeriodic;
    iTermUpdatePeriodic = NULL;
    iTermUpdatePeriodic = CPeriodic::NewL(EPriorityNormal);
    iTermUpdatePeriodic->Start(25000, 25000, TCallBack(UpdateTerminal, NULL));

    iState = EStateConnected;

    return KErrNone;
}


// MPuttyEngine::GetErrorMessage()
void CPuttyEngineImp::GetErrorMessage(TDes &aTarget) {

    if ( iConnError == NULL ) {
        aTarget.SetLength(0);
        return;
    }

    StringToDes(iConnError, aTarget);
}


// MPuttyEngine::Disconnect()
void CPuttyEngineImp::Disconnect() {

    // ARGH! FIXME! This won't really work if the connection is still open.
    // There is really no way to close a PuTTY connection forcibly except by
    // terminating the whole thing!
    assert((iState == EStateConnected) || (iState == EStateDisconnected));

    // Get rid of the timer
    delete iTermUpdatePeriodic;
    iTermUpdatePeriodic = NULL;

    // Close networking
    sk_set_watcher(NULL);
    sk_cleanup();

    iState = EStateInitialized;
    iConnError = NULL;
}


// MPuttyEngine::SetTerminalSize();
void CPuttyEngineImp::SetTerminalSize(TInt aWidth, TInt aHeight) {
    iTermWidth = aWidth;
    iTermHeight = aHeight;
    assert((iTermWidth > 1) && (iTermHeight > 1));

    if ( iState == EStateConnected ) {
        term_size(iTermHeight, iTermWidth, cfg.savelines);
    }
}


// MPuttyEngine::RePaintWindow()
void CPuttyEngineImp::RePaintWindow() {
    if ( iState != EStateConnected ) {
        return;
    }

    term_paint((Context)42, 0, 0, iTermWidth-1, iTermHeight-1);
    // term_paint only repaints if alt_pressed is true?
    // (FIXME, or rather fix terminal.c?)
    term_update();
}


// MPuttyEngine::SendKeypress
void CPuttyEngineImp::SendKeypress(TKeyCode aCode, TUint aModifiers) {
    if ( iState != EStateConnected ) {
        return;
    }

    SendKey(aCode, aModifiers);
}


// MPuttyEngine::AddRandomNoise
void CPuttyEngineImp::AddRandomNoise(const TDesC8& aNoise) {
    assert(iState != EStateNone);
    random_add_noise((void*)aNoise.Ptr(), aNoise.Length());
}


// MSocketWatcher::SocketOpened()
void CPuttyEngineImp::SocketOpened() {
    iNumSockets++;
}


// MSocketWatcher::SocketClosed()
void CPuttyEngineImp::SocketClosed() {
    iNumSockets--;
    if ( (iState == EStateConnected) && (iNumSockets == 0) ) {
        // Disconnect, signal ourselves
        iState = EStateDisconnected;
        SetActive();
        TRequestStatus *status = &iStatus;
        User::RequestComplete(status, KErrNone);
    }
}


// Fatal error (PuTTY callback)
void CPuttyEngineImp::putty_fatalbox(char *p, va_list ap) {

    char *buf = (char*) smalloc(1024);
    HBufC *des = HBufC::New(1024);
    if ( (buf == NULL) || (des == NULL) ) {
        iClient->FatalError(KOutOfMemory);
    }

    vsprintf(buf, p, ap);
    TPtr ptr = des->Des();
    StringToDes(buf, ptr);
    iClient->FatalError(ptr);
    delete des;
    sfree(buf);
}

void fatalbox(char *p, ...) {
    assert(engine);
    va_list ap = 0;
    va_start(ap, p);
    engine->putty_fatalbox(p, ap);
    va_end(ap);
}


// Fatal connection error (PuTTY callback)
void CPuttyEngineImp::putty_connection_fatal(char *p, va_list ap) {

    char *buf = (char*) smalloc(1024);
    HBufC *des = HBufC::New(1024);
    if ( (buf == NULL) || (des == NULL) ) {
        iClient->FatalError(KOutOfMemory);
    }

    vsprintf(buf, p, ap);
    TPtr ptr = des->Des();
    StringToDes(buf, ptr);
    iClient->ConnectionError(ptr);
    delete des;
    sfree(buf);
}

void connection_fatal(char *p, ...) {
    assert(engine);
    va_list ap = 0;
    va_start(ap, p);
    engine->putty_connection_fatal(p, ap);
    va_end(ap);
}


// Draw text on screen
void CPuttyEngineImp::putty_do_text(int x, int y, char *text, int len,
                                    unsigned long attr, int /*lattr*/) {

    // Make sure we have a resonable character set
    assert((attr & CSET_MASK) != ATTR_OEMCP);
    assert((attr & CSET_MASK) != ATTR_ACP);
    assert(!DIRECT_CHAR(attr));
    assert(!DIRECT_FONT(attr));

    // Convert to proper wide chars (why aren't they so already? grr)
    static TUint16 *wbuf = NULL;
    static int wlen = 0;
    int i;
    if ( wlen < len ) {
        sfree(wbuf);
        wlen = len;
        wbuf = (TUint16*) smalloc(wlen * sizeof(TUint16));
    }
    for (i = 0; i < len; i++)
        wbuf[i] = (TUint16) ((attr & CSET_MASK) + (text[i] & CHAR_MASK));

    // Handle cursor attributes
    if ( attr & TATTR_ACTCURS ) {
	attr &= ATTR_CUR_AND;
	attr ^= ATTR_CUR_XOR;
    }

    // Determine the colors to use
    TInt fgIndex = 2 * ((attr & ATTR_FGMASK) >> ATTR_FGSHIFT);
    TInt bgIndex = 2 * ((attr & ATTR_BGMASK) >> ATTR_BGSHIFT);
    if ( attr & ATTR_REVERSE ) {
        TInt tmp = fgIndex;
        fgIndex = bgIndex;
        bgIndex = tmp;
    }
    // FIXME: Bold colors
#if 0
    if ( (bold_mode == BOLD_COLOURS) && (attr & ATTR_BOLD) ) {
        fgIndex++;
    }
    if ( (bold_mode == BOLD_COLOURS) && (attr & ATTR_BLINK) ) {
        bgIndex++;
    }
#endif
    assert((fgIndex >= 0) && (bgIndex >= 0) &&
           (fgIndex < KNumColors) && (bgIndex < KNumColors));

    // Draw -- FIXME: Colors and attributes!
    TPtrC16 ptr(wbuf, len);
    iClient->DrawText(x, y, ptr, EFalse, EFalse, iPalette[fgIndex],
                      iPalette[bgIndex]);
}

void do_text(Context /*ctx*/, int x, int y, char *text, int len,
	     unsigned long attr, int lattr) {
    assert(engine);
    engine->putty_do_text(x, y, text, len, attr, lattr);
}


// Verify SSH host key (PuTTY callback)
void CPuttyEngineImp::putty_verify_ssh_host_key(
    char *host, int port, char *keytype, char *keystr, char *fingerprint) {

    // Verify key against the store
    int keystatus = verify_host_key(host, port, keytype, keystr);

    if ( keystatus == 0 ) {
        return; // matched to a previously stored key
    }

    TPtr *fpDes = CreateDes(fingerprint);
    MPuttyClient::THostKeyResponse resp = MPuttyClient::EAbadonConnection;

    // Key not verified OK, prompt the user
    if (keystatus == 2 ) {
        // Key in store, but different
        resp = iClient->DifferentHostKey(*fpDes);
    } else if ( keystatus == 1 ) {
        // Key not in store
        resp = iClient->UnknownHostKey(*fpDes);
    } else {
        assert(EFalse);
    }

    DeleteDes(fpDes);

    // React
    switch ( resp ) {
        case MPuttyClient::EAbadonConnection:
            // FIXME! There is no other safe way to close connection _NOW, and
            // we must be sure we don't send anything to the server
            User::Exit(KErrCancel);
            break;

        case MPuttyClient::EAcceptTemporarily:
            break;

        case MPuttyClient::EAcceptAndStore:
	    store_host_key(host, port, keytype, keystr);
            break;

        default:
            assert(EFalse);
    }
}

void verify_ssh_host_key(char *host, int port, char *keytype,
                         char *keystr, char *fingerprint) {
    engine->putty_verify_ssh_host_key(host, port, keytype, keystr,
                                      fingerprint);
}



// Prompt the user to accept a cipher below the warning threshold
void CPuttyEngineImp::putty_askcipher(char *ciphername, int cs) {
    
    TPtr *cipherDes = CreateDes(ciphername);
    MPuttyClient::TCipherDirection dir = MPuttyClient::EBothDirections;

    // Convert direction to client interface constant
    switch ( cs ) {
        case 0:
            dir = MPuttyClient::EBothDirections;
            break;

        case 1:
            dir = MPuttyClient::EClientToServer;
            break;

        case 2:
            dir = MPuttyClient::EServerToClient;
            break;

        default:
            assert(EFalse);
    }

    // Prompt the user, kill connection if the cipher is not acceptable
    if ( !iClient->AcceptCipher(*cipherDes, dir) ) {
        // FIXME!
        DeleteDes(cipherDes);
        User::Exit(KErrCancel);
    }

    DeleteDes(cipherDes);
}


void askcipher(char *ciphername, int cs) {
    engine->putty_askcipher(ciphername, cs);
}


void old_keyfile_warning(void)
{
    // FIXME?
    fatalbox("Unsupported old private key file format");
}


// Set a palette entry (copied pretty much from window.c)
void CPuttyEngineImp::putty_palette_set(int n, int r, int g, int b) {
    
    static const int first[21] = {
	0, 2, 4, 6, 8, 10, 12, 14,
	1, 3, 5, 7, 9, 11, 13, 15,
	16, 17, 18, 20, 22
    };

    assert((n > 0) && (n < 21));
    assert((r > 0) && (r < 256));
    assert((g > 0) && (g < 256));
    assert((b > 0) && (g < 256));
    
    iPalette[first[n]].SetRed(r);
    iPalette[first[n]].SetGreen(g);
    iPalette[first[n]].SetBlue(b);

    if ( first[n] >= 18 ) {
        iPalette[first[n]+1].SetRed(r);
        iPalette[first[n]+1].SetGreen(g);
        iPalette[first[n]+1].SetBlue(b);
    }
}

void palette_set(int n, int r, int g, int b)
{
    engine->putty_palette_set(n, r, g, b);
}


// Reset to the default palette
void CPuttyEngineImp::putty_palette_reset() {
    Mem::Copy(iPalette, iDefaultPalette, sizeof(iPalette));
}

void palette_reset(void)
{
    engine->putty_palette_reset();
}


// Draw the cursor
void CPuttyEngineImp::putty_do_cursor(int x, int y, char *text, int len,
                                      unsigned long attr, int lattr) {

    // FIXME? We currently only support a block cursor, and just draw it as
    // text with the correct colors
    if ( attr & TATTR_ACTCURS ) {
        putty_do_text(x, y, text, len, attr, lattr);
    }
}

void do_cursor(Context /*ctx*/, int x, int y, char *text, int len,
	       unsigned long attr, int lattr) {
    engine->putty_do_cursor(x, y, text, len, attr, lattr);
}


// Clean up and exit
void cleanup_exit(int code) {
    delete engine;
    User::Exit(code);
}


// Update terminal
int CPuttyEngineImp::UpdateTerminal(TAny * /*anAny*/) {
    term_out();
    term_update();
    return 0;
}


void CPuttyEngineImp::RunL() {

    assert(iStatus.Int() == KErrNone);

    if ( iState == EStateDisconnected ) {
        Disconnect();
        iClient->ConnectionClosed();
    } else {
        assert(EFalse);
    }
}


void CPuttyEngineImp::DoCancel() {
    assert(EFalse);
}



/**********************************************************
 *
 * Other PuTTY callback functions, most of these do nothing
 *
 **********************************************************/


Context get_ctx(void)
{
    return (void*) 42;
}

void free_ctx(Context /*ctx*/)
{
}


/*
 * Move the system caret. (We maintain one, even though it's
 * invisible, for the benefit of blind people: apparently some
 * helper software tracks the system caret, so we should arrange to
 * have one.)
 */
void sys_cursor(int /*x*/, int /*y*/)
{
}


void set_sbar(int /*total*/, int /*start*/, int /*page*/)
{
}


char *get_window_title(int icon)
{
    return icon ? "PuTTY.Icon" : "PuTTY";
}


/*
 * Report the window's position, for terminal reports.
 */
void get_window_pos(int *x, int *y)
{
    *x = 0;
    *y = 0;
}

/*
 * Report the window's pixel size, for terminal reports.
 */
void get_window_pixels(int *x, int *y)
{
    *x = 640;
    *y = 200;
}

/*
 * Report whether the window is iconic, for terminal reports.
 */
int is_iconic(void)
{
    return 0;
}


/*
 * Maximise or restore the window in response to a server-side
 * request.
 */
void set_zoomed(int /*zoomed*/)
{
}

/*
 * Refresh the window in response to a server-side request.
 */
void refresh_window(void)
{
}


/*
 * Move the window to the top or bottom of the z-order in response
 * to a server-side request.
 */
void set_zorder(int /*top*/)
{
}

/*
 * Move the window in response to a server-side request.
 */
void move_window(int /*x*/, int /*y*/)
{
}


/*
 * Minimise or restore the window in response to a server-side
 * request.
 */
void set_iconic(int /*iconic*/)
{
}

void request_resize(int /*w*/, int /*h*/)
{
}

/*
 * Beep.
 */
void beep(int /*mode*/)
{
}

/*
 * set or clear the "raw mouse message" mode
 */
void set_raw_mouse_mode(int /*activate*/)
{
}

void set_title(char * /*title*/)
{
}

void set_icon(char * /*title*/)
{
}



/*
 * Writes stuff to clipboard
 */
void write_clip(wchar_t * /*data*/, int /*len*/, int /*must_deselect*/)
{
}

void get_clip(wchar_t ** /*p*/, int * /*len*/)
{
}


/*
 * Translate a raw mouse button designation (LEFT, MIDDLE, RIGHT)
 * into a cooked one (SELECT, EXTEND, PASTE).
 */
Mouse_Button translate_button(Mouse_Button /*button*/)
{
    return (Mouse_Button) 0;
}


/* This function gets the actual width of a character in the normal font.
 */
int CharWidth(Context /*ctx*/, int /*uc*/) {

    return 1;
}

extern "C" void ldisc_update(int /*echo*/, int /*edit*/)
{
}

void logevent(char * /*string*/) {
}

void log_packet(int /*direction*/, int /*type*/, char * /*texttype*/,
                void * /*data*/, int /*len*/) {
}

void logtraffic(unsigned char /*c*/, int /*logmode*/)
{
}

char *do_select(RSocketS * /*skt*/, int /*startup*/)
{
    return NULL;
}
