/*    epocnet.cpp
 *
 * Symbian OS implementation of PuTTY networking support
 *
 * Mostly copied from the SyOSsh EPOC Release 5 port by Gabor Keresztfalvi,
 * some additional fixes for the current Symbian port by Petteri Kangaslampi.
 * I/O server stuff included in the SyOSsh version removed.
 * Originally baset on winnet.c in the PuTTY distribution.
 *
 * [No copyright messages in the original, assumed to be copyrighted by
 *  Gabor Keresztfavli and available through the PuTTY license]
 *
 * Portions copyright 2002 Petteri Kangaslampi
 *
 * See license.txt for full copyright and license information.
*/

/*
 * SymbianOS (EPOC) networking abstraction. Version 2. Uses ActiveObjects...
 * FIXME:
 * - On MARM lookup of a valid IP number fails if having DNS server?
 */

/* #define IPV6 1 */

#include <e32std.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>

#ifdef __cplusplus
extern "C" {
#endif
#define DEFINE_PLUG_METHOD_MACROS
#include "putty.h"
#include "network.h"
#include "tree234.h"
#ifdef __cplusplus
}
#endif

#include "epocnet.h"
#include "charutil.h"


struct Socket_tag {
    struct socket_function_table *fn;
    /* the above variable absolutely *must* be the first in this structure */
    char *error;
    RSocketS *s;
    Plug plug;
    void *private_ptr;
    bufchain output_data;
    int connected;
    int writable;
    int frozen; /* this causes readability notifications to be ignored */
    int frozen_readable; /* this means we missed at least one readability
			  * notification while we were frozen */
    int localhost_only;		       /* for listening sockets */
    int listening;		       /* for listening sockets */
    RSocketS *accepted;
    char oobdata[1];
    int sending_oob;
    int oobinline;
    int pending_error;		       /* in case send() returns error */
    int killing;		/* In case we have data to send... */
};

/*
 * We used to typedef struct Socket_tag *Socket.
 *
 * Since we have made the networking abstraction slightly more
 * abstract, Socket no longer means a tcp socket (it could mean
 * an ssl socket).  So now we must use Actual_Socket when we know
 * we are talking about a tcp socket.
 */
typedef struct Socket_tag *Actual_Socket;

struct SockAddr_tag {
    char *error;
    TUint family;
    TInetAddr *address;	       /* Address IPv4 style. */
};

static tree234 *sktree;
static RSocketServ socketserv;
static int tha_global_id=1;
static TBuf8<20480> oobuffer;
static MSocketWatcher *watcher;

static int cmpfortree(void *av, void *bv)
{
    Actual_Socket a = (Actual_Socket) av, b = (Actual_Socket) bv;
    unsigned long as = (unsigned long) a->s->id, bs = (unsigned long) b->s->id;
    if (as < bs)
	return -1;
    if (as > bs)
	return +1;
    return 0;
}

static int cmpforsearch(void *av, void *bv)
{
    Actual_Socket b = (Actual_Socket) bv;
    unsigned long as = (unsigned long) ((RSocketS *)av)->id, bs = (unsigned long) b->s->id;
    if (as < bs)
	return -1;
    if (as > bs)
	return +1;
    return 0;
}

void sk_init(void)
{
    TInt err;
    sktree = newtree234(cmpfortree);

    err=socketserv.Connect(2*MAXSOCKETS);
    User::LeaveIfError(err);
}

void sk_cleanup(void)
{
    Socket s;
    int i;

    if (sktree) {
	for (i = 0; (s = (Socket)index234(sktree, i)) != NULL; i++) {
            sk_close(s);
	}
        
        socketserv.Close();

        freetree234(sktree);
        sktree = NULL;
    }
}

// Set watcher object. Set to NULL to disable.
void sk_set_watcher(MSocketWatcher *aWatcher) {
    watcher = aWatcher;
}

char *sock_error_string(TInt error)
{
    switch (error) {
	case KErrNone:
			return("No error. (None)");
	case KErrNotFound:
			return("Item not found. (NotFound)");
	case KErrGeneral:
			return("Uncategorized error. (General)");
	case KErrCancel:
			return("Operation cancelled. (Cancel)");
	case KErrNoMemory:
			return("A memory allocation failed. (NoMemory)");
	case KErrNotSupported:
			return("A function is not supported in a given context. (NotSupported)");
	case KErrArgument:
			return("An argument is out of range. (Argument)");
	case KErrBadHandle:
			return("Bad handle. (BadHandle)");
	case KErrOverflow:
			return("Overflow. (Overflow)");
	case KErrUnderflow:
			return("Underflow. (Underflow)");
	case KErrAlreadyExists:
			return("The resource already exists. (AlreadyExists)");
	case KErrPathNotFound:
			return("In the context of file operations, the path was not found. (PathNotFound)");
	case KErrDied:
			return("A handle refers to a thread which has died (Died)");
	case KErrInUse:
			return("A requested resource is already in use. (InUse)");
	case KErrServerTerminated:
			return("A client/server operation cannot execute, because the server has terminated. (ServerTerminated)");
	case KErrServerBusy:
			return("A client/server operation cannot execute, because the server is busy. (ServerBusy)");
	case KErrNotReady:
			return("Resource not ready. Not initialized, or has no power. (NotReady)");
	case KErrUnknown:
			return("A device is of unknown type. (Unknown)");
	case KErrCorrupt:
			return("Corrupted. (Corrupt)");
	case KErrAccessDenied:
			return("Access denied. (AccessDenied)");
	case KErrLocked:
			return("The operation cannot be performed, because the resource is locked. (Locked)");
	case KErrWrite:
			return("During a file write operation, not all the data could be written. (Write)");
	case KErrDisMounted:
			return("A volume which was to be used for a file system operation has been dismounted. (DisMounted)");
	case KErrEof:
			return("End of file has been reached. (Eof)");
	case KErrDiskFull:
			return("A write operation could not complete, because the disk was full. (DiskFull)");
	case KErrBadDriver:
			return("A driver DLL was of the wrong type. (BadDriver)");
	case KErrBadName:
			return("Name did not conform with the required syntax. (BadName)");
	case KErrCommsLineFail:
			return("The communication line failed. (CommsLineFail)");
	case KErrCommsFrame:
			return("A frame error occurred in a communications operation. (CommsFrame)");
	case KErrCommsOverrun:
			return("An overrun was detected by a communications driver. (CommsOverrun)");
	case KErrCommsParity:
			return("A parity error occurred in communications. (CommsParity)");
	case KErrTimedOut:
			return("An operation timed out. (TimedOut)");
	case KErrCouldNotConnect:
			return("A session could not connect. (CouldNotConnect)");
	case KErrCouldNotDisconnect:
			return("A session could not disconnect. (CouldNotDisconnect)");
	case KErrDisconnected:
			return("The required session was disconnected. (Disconnected)");
	case KErrBadLibraryEntryPoint:
			return("A library entry point was not of the required type. (BadLibraryEntryPoint)");
	case KErrBadDescriptor:
			return("A non-descriptor parameter was passed. (BadDescriptor)");
	case KErrAbort:
			return("An operation was aborted (Abort)");
	case KErrTooBig:
			return("A number was too big (TooBig)");
	case KErrDivideByZero:
			return("A divide-by-zero operation was attempted. (DivideByZero)");
	case KErrBadPower:
			return("Insufficient power was available to complete an operation. (BadPower)");
	case KErrWouldBlock:
			return("Network error: Resource temporarily unavailable (WouldBlock)");
	case KErrNetUnreach:
			return("Network unreachable. (NetUnreach)");
	case KErrHostUnreach:
			return("Host unreachable. (HostUnreach)");
	case KErrNoProtocolOpt:
			return("No such protocol option. (NoProtocolOpt)");
	case KErrUrgentData:
			return("Urgent data arrived. (UrgentData)");
	case KInvalSocket:
			return("Got NULL sokcet.");
	default:
			return("Unknown network error");
    }
}

SockAddr sk_namelookup(char *host, char **canonicalname)
{
    SockAddr ret = (SockAddr) smalloc(sizeof(struct SockAddr_tag));
    RHostResolver hres;
    TNameEntry nameres;
    TNameRecord namerec;
    THostName thehostname;
    TInt err;

    memset(ret, 0, sizeof(struct SockAddr_tag));
    ret->family = KAfInet;

    ret->address=new TInetAddr();
    if (ret->address==NULL)
    {
	    ret->error="Not enough memory to allocate TInetAddr.";
	    return ret;
    }
    err=hres.Open(socketserv, KAfInet, KProtocolInetUdp);
    if (err!=KErrNone)
    {
	    ret->error=sock_error_string(err);
	    delete ret->address;
	    ret->address=NULL;
	    return ret;
    }

    TPtr *hostDes = CreateDes(host);
    err=hres.GetByName(*hostDes,nameres);
    DeleteDes(hostDes);
    if (err!=KErrNone)
    {
	    ret->error=sock_error_string(err);
	    delete ret->address;
	    ret->address=NULL;
	    hres.Close();
	    return ret;
    }
    namerec=nameres();
    ret->address->SetAddress(((TInetAddr)namerec.iAddr).Address());

    err=hres.GetByAddress(namerec.iAddr,nameres);	// Should compile as exedll project, or this fails on WINS...
    if ( err == KErrNone )
    {
        namerec=nameres();
        thehostname=namerec.iName;
        *canonicalname = (char *)smalloc(1+thehostname.Length());
        DesToString(thehostname, *canonicalname);
    }
    else
    {
        // The call can often fail if the server has no valid reverse. This
        // happens, for example, if the name is a numeric address.
        *canonicalname = (char*) smalloc(strlen(host)+1);
        strcpy(*canonicalname, host);
    }

    hres.Close();
    return ret;
}

void sk_getaddr(SockAddr addr, char *buf, int buflen)
{
    TBuf<40> ip;
    int thesize;

    addr->address->Output(ip);
    thesize = buflen<ip.Size()+1 ? buflen-1 : ip.Size();
    strncpy(buf, (char *)ip.Ptr(), thesize);
    buf[thesize]='\0';
}

int sk_addrtype(SockAddr addr)
{
    if ( addr->family == KAfInet )
        return AF_INET;
    return 0;
}

void sk_addrcopy(SockAddr addr, char *buf)
{
    TUint32 a = addr->address->Address();
    TUint32 n = htonl(a);
    memcpy(buf, &n, 4);
}

void sk_addr_free(SockAddr addr)
{
    if (addr->address) delete addr->address;
    sfree(addr);
}

static Plug sk_tcp_plug(Socket sock, Plug p)
{
    Actual_Socket s = (Actual_Socket) sock;
    Plug ret = s->plug;
    if (p)
	s->plug = p;
    return ret;
}

static void sk_tcp_flush(Socket /*s*/)
{
    /*
     * We send data to the socket as soon as we can anyway,
     * so we don't need to do anything here.  :-)
     */
}

static void sk_tcp_close(Socket s);
static int sk_tcp_write(Socket s, char *data, int len);
static int sk_tcp_write_oob(Socket s, char *data, int len);
static void sk_tcp_set_private_ptr(Socket s, void *ptr);
static void *sk_tcp_get_private_ptr(Socket s);
static void sk_tcp_set_frozen(Socket s, int is_frozen);
static char *sk_tcp_socket_error(Socket s);

extern char *do_select(RSocketS *skt, int startup);

Socket sk_register(void *sock, Plug plug)
{
    static struct socket_function_table fn_table = {
	sk_tcp_plug,
	sk_tcp_close,
	sk_tcp_write,
	sk_tcp_write_oob,
	sk_tcp_flush,
	sk_tcp_set_private_ptr,
	sk_tcp_get_private_ptr,
	sk_tcp_set_frozen,
	sk_tcp_socket_error
    };

    char *errstr;
    Actual_Socket ret;
    TInt err;

    /*
     * Create Socket structure.
     */
    ret = (Actual_Socket)smalloc(sizeof(struct Socket_tag));
    ret->fn = &fn_table;
    ret->error = NULL;
    ret->plug = plug;
    bufchain_init(&ret->output_data);
    ret->writable = 1;		       /* to start with */
    ret->sending_oob = 0;
    ret->frozen = 1;
    ret->frozen_readable = 0;
    ret->localhost_only = 0;	       /* unused, but best init anyway */
    ret->listening = 0;		       /* unused, but best init anyway */
    ret->pending_error = 0;
    ret->accepted=NULL;
    ret->killing=0;

    ret->s = (RSocketS *)sock;

    if (ret->s == INVALID_SOCKET) {
	ret->error = sock_error_string(KInvalSocket);
	return (Socket) ret;
    }

    ret->s->estat=KRequestPending;
    ret->s->r=NULL;
    ret->s->s=NULL;
    ret->s->a=NULL;

    ret->oobinline = 0;
    errstr = do_select(ret->s, 1);
    if (errstr) {
	ret->error = errstr;
	return (Socket) ret;
    }

    TRAP(err,ret->s->r = CRecver::NewL((Socket)ret));
    ret->s->r->StartRecver();
    TRAP(err,ret->s->s = CSender::NewL((Socket)ret));

    add234(sktree, ret);

    return (Socket) ret;
}

Socket sk_new(SockAddr addr, int port, int privport, int oobinline,
	      int nodelay, Plug plug)
{
    static struct socket_function_table fn_table = {
	sk_tcp_plug,
	sk_tcp_close,
	sk_tcp_write,
	sk_tcp_write_oob,
	sk_tcp_flush,
	sk_tcp_set_private_ptr,
	sk_tcp_get_private_ptr,
	sk_tcp_set_frozen,
	sk_tcp_socket_error
    };

    TInt err;
    TInetAddr locaddr;
    TRequestStatus cstat;
    char *errstr;
    Actual_Socket ret;
    short localport;

    /*
     * Create Socket structure.
     */
    ret = (Actual_Socket) smalloc(sizeof(struct Socket_tag));
    ret->fn = &fn_table;
    ret->error = NULL;
    ret->plug = plug;
    bufchain_init(&ret->output_data);
    ret->connected = 0;		       /* to start with */
    ret->writable = 0;		       /* to start with */
    ret->sending_oob = 0;
    ret->frozen = 0;
    ret->frozen_readable = 0;
    ret->localhost_only = 0;	       /* unused, but best init anyway */
    ret->listening = 0;		       /* unused, but best init anyway */
    ret->pending_error = 0;
    ret->accepted=NULL;
    ret->killing=0;

    /*
     * Open socket.
     */
    ret->s = new RSocketS;
    if (ret->s == NULL) {
	ret->error = "Not enough memory to allocate RSocketS";
	return (Socket) ret;
    }
    ret->s->id=tha_global_id++;
    ret->s->estat=KRequestPending;
    ret->s->r=NULL;
    ret->s->s=NULL;
    ret->s->a=NULL;

    if ( watcher ) {
        watcher->SocketOpened();
    }

    err=ret->s->Open(socketserv, addr->family, KSockStream, KUndefinedProtocol);
    if (err!=KErrNone) {
	    ret->error = sock_error_string(err);
	    return (Socket) ret;
    }

    ret->oobinline = oobinline;
    if (oobinline) {
	err=ret->s->SetOpt(KSoTcpOobInline, KSolInetTcp, 1);
	if (err!=KErrNone) {
	    ret->error = sock_error_string(err);
	    return (Socket) ret;
        }
    }

    if (nodelay) {
	err=ret->s->SetOpt(KSoTcpNoDelay, KSolInetTcp, 1);
	if (err!=KErrNone) {
	    ret->error = sock_error_string(err);
	    return (Socket) ret;
        }
    }

    /*
     * Bind to local address.
     */
    if (privport)
	localport = 1023;	       /* count from 1023 downwards */
    else
	localport = 0;		       /* just use port 0 (ie winsock picks) */

    /* Loop round trying to bind */
    while (localport >= 0) {

	locaddr = TInetAddr(KInetAddrAny, localport);
	err = ret->s->Bind(locaddr);
	if (err == KErrNone) break;		       /* done */
	else {
	    if (err != KErrInUse) break; /* failed, for a bad reason */
	}

	if (localport == 0)
	    break;		       /* we're only looping once */
	localport--;
	if (localport == 0)
	    break;		       /* we might have got to the end */
    }

    if (err!=KErrNone) {
	ret->error = sock_error_string(err);
	return (Socket) ret;
    }

    /*
     * Connect to remote address.
     */

    TInetAddr tempaddr(addr->address->Address(), port);
    ret->s->Connect(tempaddr, cstat);
    // FIXME: This is potentially unsafe -- it could eat events from the
    // active scheduler
    User::WaitForRequest(cstat);	// We block it here to ease our problem in the active objects.
    if ( cstat != KErrNone ) {
	    ret->error = sock_error_string(cstat.Int());
	    return (Socket) ret;
	}

    ret->connected = 1;
    ret->writable = 1;

    errstr = do_select(ret->s, 1);
    if (errstr) {
	ret->error = errstr;
	return (Socket) ret;
    }

    TRAP(err, ret->s->r = CRecver::NewL((Socket)ret));
    ret->s->r->StartRecver();
    TRAP(err,ret->s->s = CSender::NewL((Socket)ret));

    add234(sktree, ret);

    return (Socket) ret;
}

Socket sk_newlistener(int port, Plug plug, int local_host_only)
{
    static struct socket_function_table fn_table = {
	sk_tcp_plug,
	sk_tcp_close,
	sk_tcp_write,
	sk_tcp_write_oob,
	sk_tcp_flush,
	sk_tcp_set_private_ptr,
	sk_tcp_get_private_ptr,
	sk_tcp_set_frozen,
	sk_tcp_socket_error
    };

    TInt err;
    TInetAddr locaddr;
//    char *errstr;
    Actual_Socket ret;

    /*
     * Create Socket structure.
     */
    ret = (Actual_Socket)smalloc(sizeof(struct Socket_tag));
    ret->fn = &fn_table;
    ret->error = NULL;
    ret->plug = plug;
    bufchain_init(&ret->output_data);
    ret->writable = 0;		       /* to start with */
    ret->sending_oob = 0;
    ret->frozen = 0;
    ret->frozen_readable = 0;
    ret->localhost_only = local_host_only;
    ret->listening = 1;
    ret->pending_error = 0;
    ret->accepted=NULL;
    ret->killing=0;

    /*
     * Open socket.
     */
    ret->s = new RSocketS;
    if (ret->s == NULL) {
	ret->error = "Not enough memory to allocate RSocketS";
	return (Socket) ret;
    }
    ret->s->id=tha_global_id++;
    ret->s->estat=KRequestPending;
    ret->s->r=NULL;
    ret->s->s=NULL;
    ret->s->a=NULL;

    if ( watcher ) {
        watcher->SocketOpened();
    }

    err=ret->s->Open(socketserv, KAfInet, KSockStream, KUndefinedProtocol);
    if (err!=KErrNone) {
	    ret->error = sock_error_string(err);
	    return (Socket) ret;
    }

    ret->oobinline = 0;

    ret->s->SetOpt(KSoReuseAddr, KSolInetIp, 1);

    locaddr = TInetAddr(local_host_only ? KInetAddrLoop : KInetAddrAny, port);
    err = ret->s->Bind(locaddr);
    if (err) {
	ret->error = sock_error_string(err);
	return (Socket) ret;
    }


    err = ret->s->Listen(3);
    if (err!=KErrNone) {
        ret->s->Close();
	ret->error = sock_error_string(err);
	return (Socket) ret;
    }

    TRAP(err,ret->s->a = CAcceptor::NewL((Socket)ret));
    ret->s->a->StartAcceptor();

    add234(sktree, ret);

    return (Socket) ret;
}

static void sk_tcp_close(Socket sock)
{
    Actual_Socket s = (Actual_Socket) sock;

    if (bufchain_size(&(s->output_data))>0 && s->killing==0)
    {
	    s->killing=1;
	    return;
    }
    if ( watcher ) {
        watcher->SocketClosed();
    }
    del234(sktree, s);
    do_select(s->s, 0);
    if (s->accepted!=NULL) delete s->accepted;
    if (s->s->r!=NULL) delete s->s->r;
    if (s->s->s!=NULL) delete s->s->s;
    if (s->s->a!=NULL) delete s->s->a;
    s->s->Close();
    delete s->s;
    sfree(s);
}

static int sk_tcp_write(Socket sock, char *buf, int len)
{
    Actual_Socket s = (Actual_Socket) sock;

    /*
     * Add the data to the buffer list on the socket.
     */
    bufchain_add(&s->output_data, buf, len);

    /*
     * Now try sending from the start of the buffer list.
     */
    if (s->writable) s->s->s->Sendit();

    return bufchain_size(&s->output_data);
}

static int sk_tcp_write_oob(Socket sock, char *buf, int len)
{
    Actual_Socket s = (Actual_Socket) sock;

    /*
     * Replace the buffer list on the socket with the data.
     */
    bufchain_clear(&s->output_data);
    __ASSERT_ALWAYS(len <= (int)sizeof(s->oobdata),User::Leave(KErrOverflow));
    memcpy(s->oobdata, buf, len);
    s->sending_oob = len;

    /*
     * Now try sending from the start of the buffer list.
     */
    if (s->writable) s->s->s->Sendit();

    return s->sending_oob;
}

int select_result(RSocketS *wParam, int /*lParam*/)
{
    // FIXME: I guess this function isn't used for anything?
//    TInt err;
    Actual_Socket s;
//    TInt atmark;

    s = (Actual_Socket) find234(sktree, (void *) wParam, cmpforsearch);
    if (!s) return 1;		       /* boggle */
    return 0;
}

/*
 * Deal with socket errors detected in try_send().
 * FIXME: This can be done in the AO CSender.RunL?
 */
void net_pending_errors(void)
{
    int i;
    Actual_Socket s;

    /*
     * This might be a fiddly business, because it's just possible
     * that handling a pending error on one socket might cause
     * others to be closed. (I can't think of any reason this might
     * happen in current SSH implementation, but to maintain
     * generality of this network layer I'll assume the worst.)
     * 
     * So what we'll do is search the socket list for _one_ socket
     * with a pending error, and then handle it, and then search
     * the list again _from the beginning_. Repeat until we make a
     * pass with no socket errors present. That way we are
     * protected against the socket list changing under our feet.
     */

    do {
	for (i = 0; (s = (Actual_Socket)index234(sktree, i)) != NULL; i++) {
	    if (s->pending_error) {
		/*
		 * An error has occurred on this socket. Pass it to the
		 * plug.
		 */
		plug_closing(s->plug,
			     sock_error_string(s->pending_error),
			     s->pending_error, 0);
		break;
	    }
	}
    } while (s);
}

/*
 * Each socket abstraction contains a `void *' private field in
 * which the client can keep state.
 */
static void sk_tcp_set_private_ptr(Socket sock, void *ptr)
{
    Actual_Socket s = (Actual_Socket) sock;
    s->private_ptr = ptr;
}

static void *sk_tcp_get_private_ptr(Socket sock)
{
    Actual_Socket s = (Actual_Socket) sock;
    return s->private_ptr;
}

/*
 * Special error values are returned from sk_namelookup and sk_new
 * if there's a problem. These functions extract an error message,
 * or return NULL if there's no problem.
 */
char *sk_addr_error(SockAddr addr)
{
    return addr->error;
}
static char *sk_tcp_socket_error(Socket sock)
{
    Actual_Socket s = (Actual_Socket) sock;
    return s->error;
}

static void sk_tcp_set_frozen(Socket sock, int is_frozen)
{
    Actual_Socket s = (Actual_Socket) sock;

    if (s->frozen == is_frozen)	return;
    s->frozen = is_frozen;
    if (!is_frozen && s->frozen_readable)
    {
	    s->s->r->Reissue();
    }
    s->frozen_readable = 0;
}

/*
 * For Plink: enumerate all sockets currently active.
 */
SOCKET first_socket(int *state)
{
    Actual_Socket s;
    *state = 0;
    s = (Actual_Socket) index234(sktree, (*state)++);
    return s ? s->s : INVALID_SOCKET;
}

SOCKET next_socket(int *state)
{
    Actual_Socket s = (Actual_Socket) index234(sktree, (*state)++);
    return s ? s->s : INVALID_SOCKET;
}



// Get number of active sockets
int sk_num_active_sockets(void) {
    if ( !sktree ) {
        return 0;
    }
    return count234(sktree);
}

/* The ActiveObject approach... */
/* CRecver implementation */
CRecver* CRecver::NewL(Socket p)
{
	CRecver *self=new (ELeave) CRecver;
	CleanupStack::PushL(self);
	self->ConstructL(p);
	CleanupStack::Pop();
	return(self);
}

CRecver::CRecver() : CActive(CActive::EPriorityStandard),readlen(2048)
{
}

void CRecver::ConstructL(Socket p)
{
	parent=p;
	CActiveScheduler::Add(this);
}

CRecver::~CRecver()
{
	Cancel();
}

void CRecver::StartRecver()
{
	Actual_Socket s=(Actual_Socket)parent;

	s->s->RecvOneOrMore(readbuf, 0, iStatus, readlen);
	SetActive();
}

void CRecver::Reissue()
{
	RunL();
}

void CRecver::RunL()
{
    TRAPD(error, DoRunL());
    if ( error != KErrNone )
    {
        _LIT(KPanic, "CRecver");
        User::Panic(KPanic, error);
    }
}

void CRecver::DoRunL()
{
	Actual_Socket s=(Actual_Socket)parent;
	TInt atmark, err;

	err = iStatus.Int();
        
	if ( (err == KErrEof) || (readlen() == 0) )
	{
            // Remote end closed connection
	    plug_closing(s->plug, NULL, 0, 0);
	    return;
	}
	if ( err != KErrNone)
	{
	    plug_closing(s->plug, sock_error_string(err), err, 0);
	    return;
	}
        
	if (s->killing) return;
	if (s->frozen)
	{
		s->frozen_readable=1;
		return;
	}
	/*
	 * We have received data on the socket. For an oobinline
	 * socket, this might be data _before_ an urgent pointer,
	 * in which case we send it to the back end with type==1
	 * (data prior to urgent).
	 */
	if (s->oobinline) {
	    err=s->s->GetOpt(KSoTcpRcvAtMark, KSolInetTcp, atmark);
	    if (err!=KErrNone) atmark=1;
	    if (atmark) atmark=0;
	    else atmark=1;
	    /*
	     * This is just strange for me... Now how should this work???  +G
	     * 
	     * If it does nothing, we get atmark==1,
	     * which is equivalent to `no OOB pending', so the
	     * effect will be to non-OOB-ify any OOB data.
	     */
	} else
	    atmark = 1;

	noise_ultralight(readbuf.Size());
	if (plug_receive(s->plug, atmark ? 0 : 1, (char *)readbuf.Ptr(), readbuf.Size()))
		StartRecver();
	else
	{
//		CActiveScheduler::Stop();
	}
}

void CRecver::DoCancel()
{
	Actual_Socket s=(Actual_Socket)parent;

	s->s->CancelAll();
}

/* CSender implementation */
CSender* CSender::NewL(Socket p)
{
	CSender *self=new (ELeave) CSender;
	CleanupStack::PushL(self);
	self->ConstructL(p);
	CleanupStack::Pop();
	return(self);
}

CSender::CSender() : CActive(CActive::EPriorityStandard)
{
}

void CSender::ConstructL(Socket p)
{
	parent=p;
	CActiveScheduler::Add(this);
}

CSender::~CSender()
{
	Cancel();
}

void CSender::Sendit()
{
	Actual_Socket s=(Actual_Socket)parent;
	void *data;
	TInt len;
	TUint urgentflag;

	if (s->sending_oob || bufchain_size(&(s->output_data)) > 0)
	{
		if (s->sending_oob) {
			urgentflag = KSockWriteUrgent;
			len = s->sending_oob;
			data = &(s->oobdata);
		} else {
			urgentflag = 0;
			bufchain_prefix(&(s->output_data), &data, &len);
		}

		writebuf.Set((TUint8 *)data, len);
		s->s->Send(writebuf, urgentflag, iStatus);
		s->writable = 0;
		SetActive();
	}
}

void CSender::RunL()
{
    TRAPD(error, DoRunL());
    if ( error != KErrNone )
    {
        _LIT(KPanic, "CSender");
        User::Panic(KPanic, error);        
    }
}

void CSender::DoRunL()
{
	Actual_Socket s=(Actual_Socket)parent;

	if (iStatus == KErrBadHandle || iStatus == KErrDisconnected || 
			iStatus == KErrEof)
	{
		s->pending_error = iStatus.Int();
		return;
	}
	else if (iStatus != KErrNone)
	{
		fatalbox(sock_error_string(iStatus.Int()));
	}
	if (s->sending_oob) s->sending_oob = 0;
	else bufchain_consume(&(s->output_data), writebuf.Size());
	if (!s->killing) plug_sent(s->plug, s->sending_oob + bufchain_size(&(s->output_data)));
	s->writable=1;
	if (s->sending_oob || bufchain_size(&(s->output_data)) > 0) Sendit();
	else if (s->killing) sk_tcp_close((Socket)s);
}

void CSender::DoCancel()
{
	Actual_Socket s=(Actual_Socket)parent;

	s->s->CancelAll();
}

/* CAcceptor implementation */
CAcceptor* CAcceptor::NewL(Socket p)
{
	CAcceptor *self=new (ELeave) CAcceptor;
	CleanupStack::PushL(self);
	self->ConstructL(p);
	CleanupStack::Pop();
	return(self);
}

CAcceptor::CAcceptor() : CActive(CActive::EPriorityStandard)
{
}

void CAcceptor::ConstructL(Socket p)
{
	parent=p;
	CActiveScheduler::Add(this);
}

CAcceptor::~CAcceptor()
{
	Cancel();
}

void CAcceptor::StartAcceptor()
{
	Actual_Socket listener=(Actual_Socket)parent;
	TInt err;

	listener->accepted = new (ELeave) RSocketS;

	err=listener->accepted->Open(socketserv);
	if (err!=KErrNone)
	{
		delete listener->accepted;
		listener->accepted=NULL;
		User::LeaveIfError(err);
	}
	listener->accepted->id=tha_global_id++;
	listener->s->Accept(*(listener->accepted), iStatus);
	SetActive();
}

void CAcceptor::RunL()
{
    TRAPD(error, DoRunL());
    if ( error != KErrNone )
    {
        _LIT(KPanic, "CAcceptor");
        User::Panic(KPanic, error);
    }
}

void CAcceptor::DoRunL()
{
	Actual_Socket s=(Actual_Socket)parent;
	TInetAddr t;

	if (iStatus!=KErrNone)
	{
		delete s->accepted;
		s->accepted=NULL;
		StartAcceptor();
		return;
	}
	s->accepted->RemoteName(t);
	if (s->localhost_only && t.Address() != KInetAddrLoop)
	{
		s->accepted->Close();
		delete s->accepted;
		s->accepted=NULL;
	}
	else if (plug_accepting(s->plug, (void*)s->accepted))
	{
		s->accepted->Close();
		delete s->accepted;
		s->accepted=NULL;
	}
	StartAcceptor();
}

void CAcceptor::DoCancel()
{
	Actual_Socket s=(Actual_Socket)parent;

	s->s->CancelAll();
	if (s->accepted!=NULL) delete s->accepted;
}
