/*    epocstore.cpp
 *
 * Symbian OS implementation of PuTTY settings storage interface
 *
 * Mostly copied from the SyOSsh EPOC Release 5 port by Gabor Keresztfalvi,
 * some additional fixes for the current Symbian port by Petteri Kangaslampi.
 * Originally based on winstore.c in the PuTTY distribution.
 *
 * [No copyright messages in the original, assumed to be copyrighted by
 *  Gabor Keresztfavli and available through the PuTTY license]
 *
 * Portions copyright 2002 Petteri Kangaslampi
 *
 * See license.txt for full copyright and license information.
*/

/*
 * epocstore.cpp: EPOC-specific implementation of the interface
 * defined in storage.h.
 * FIXME: Partially this is a stub only...
 */

#include <e32base.h>
#include <f32file.h>

#include <stdio.h>
#include <stdlib.h>
extern "C" {
#include "putty.h"
#include "storage.h"
}
#include <assert.h>


// Static bits (ugly, but the API gives us no choice)
static TFileName dataPath;
static RFs fs;

_LIT(KRandomFile,"random.dat");
_LIT(KHostKeysFile,"hostkeys.dat");
_LIT(KTmpHostKeysFile, "hostkeys.tmp");

static const char hex[17] = "0123456789ABCDEF";


void epoc_store_init(const TDesC &aDataPath) {
    dataPath = aDataPath;
    assert(dataPath.Length() > 1);
    // Make sure the path ends with a backslash
    if ( dataPath[dataPath.Length()-1] != '\\' ) {
        dataPath.Append('\\');
    }
    TInt err = fs.Connect();
    assert(err == KErrNone);
}


void epoc_store_free() {
    fs.Close();
}


/**
 * An utility class for reading files one line at a time. Handles all
 * related buffering and line feed conversion issues.
 */
class CLineReader : public CBase {

public:
    /** 
     * Creates a new CLineReader object and pushes it to the cleanup stack.
     * 
     * @param aFileName The file to read
     * @param aFs       The file server session to use
     * 
     * @return A new CLineReader object
     */
    static CLineReader *NewLC(const TDesC &aFileName, RFs &aFs);

    /** 
     * Destructor.
     */
    ~CLineReader();

    /** 
     * Reads a line from the input file. Empty lines are discarded,
     * and the cr/lf characters are removed. Returns an empty
     * descriptor if called at the end of the file.
     * 
     * @return The line that was read, or an empty descriptor if at EOF.
     */
    TPtrC8 &ReadLineL();

private:
    CLineReader();
    void ConstructL(const TDesC &aFileName, RFs &aFs);
    TBool ReadDataL();

    TUint8 *iBuffer;
    TUint iBufSize;
    TUint iDataLength;
    TUint iLineEnd;
    TUint iLineStart;
    TPtrC8 iThisLine;
    RFs *iFs;
    RFile iFile;
    TBool iFileOpen;
};


// Builds a new line reader object
CLineReader *CLineReader::NewLC(const TDesC &aFileName, RFs &aFs) {
    CLineReader *self = new (ELeave) CLineReader;
    CleanupStack::PushL(self);
    self->ConstructL(aFileName, aFs);
    return self;
}


CLineReader::CLineReader() : iThisLine(NULL, 0) {
    iFileOpen = EFalse;    
}

void CLineReader::ConstructL(const TDesC &aFileName, RFs &aFs) {
    iFs = &aFs;

    User::LeaveIfError(iFile.Open(*iFs, aFileName, EFileShareReadersOnly));
    iFileOpen = ETrue;
    iBufSize = 512;
    iDataLength = 0;
    iLineEnd = 0;
    iBuffer = new (ELeave) TUint8[iBufSize];
}


CLineReader::~CLineReader() {
    if ( iFileOpen ) {
        iFile.Close();
    }
    delete iBuffer;
}


// Reads a line from the file. Empty lines are discarded, and the cr/lf
// characters are removed. Returns an empty descriptor if at the end of the
// file.
TPtrC8 &CLineReader::ReadLineL() {

    // Discard previous line
    if ( iLineEnd > 0 ) {
        assert(iLineEnd <= iDataLength);
        if ( iLineEnd < iDataLength ) {
            Mem::Copy(&iBuffer[0], &iBuffer[iLineEnd], iDataLength - iLineEnd);
            iDataLength -= iLineEnd;
        }
        iLineEnd = 0;
    }
    iLineStart = 0;

    // Find where the next line starts, skipping any linefeed characters
    TBool startFound = EFalse;
    while ( !startFound ) {
        // Read more data if necessary
        if ( iLineStart >= iDataLength ) {
            if ( !ReadDataL() ) {
                // EOF
                iThisLine.Set(NULL, 0);
                return iThisLine;
            }
        }

        // Does the line start here?
        if ( (iBuffer[iLineStart] != '\r') && (iBuffer[iLineStart] != '\n') ) {
            startFound = ETrue;
        } else {
            iLineStart++;
        }
    }

    // Find where the line ends
    iLineEnd = iLineStart;
    TBool endFound = EFalse;
    while ( !endFound ) {
        // Read more data if necessary
        if ( iLineEnd >= iDataLength ) {
            if ( !ReadDataL() ) {
                // EOF -- use last line if we have one
                if ( iLineEnd > iLineStart ) {
                    iThisLine.Set(&iBuffer[iLineStart], iLineEnd - iLineStart);
                } else {
                    iThisLine.Set(NULL, 0);
                }
                return iThisLine;
            }
        }

        // Does the line end here?
        if ( (iBuffer[iLineEnd] == '\r') || (iBuffer[iLineEnd] == '\n') ) {
            endFound = ETrue;
        } else {
            iLineEnd++;
        }        
    }

    // OK, we have the line
    iThisLine.Set(&iBuffer[iLineStart], iLineEnd - iLineStart);
    return iThisLine;
}


// Internal: Reads more data to the internal buffer. Returns EFalse if at the
// EOF and couldn't read more.
TBool CLineReader::ReadDataL() {

    // Need a bigger buffer?
    if ( iDataLength >= iBufSize ) {
        TUint newSize = 2 * iBufSize;
        TUint8 *newBuf = new (ELeave) TUint8[newSize];
        Mem::Copy(newBuf, iBuffer, iBufSize);
        delete [] iBuffer;
        iBuffer = newBuf;
        iBufSize = newSize;
    }

    // Read some data
    assert(iBufSize > iDataLength);
    assert(iFileOpen);
    TPtr8 ptr(&iBuffer[iDataLength], iBufSize - iDataLength);
    User::LeaveIfError(iFile.Read(ptr));
    if ( ptr.Length() == 0 ) {
        return EFalse;
    }
    iDataLength += ptr.Length();

    return ETrue;
}




#if 0  // unused for the time being
static void mungestr(char *in, char *out)
{
    int candot = 0;

    while (*in) {
	if (*in == ' ' || *in == '\\' || *in == '*' || *in == '?' ||
	    *in == '%' || *in < ' ' || *in > '~' || (*in == '.'
						     && !candot)) {
	    *out++ = '%';
	    *out++ = hex[((unsigned char) *in) >> 4];
	    *out++ = hex[((unsigned char) *in) & 15];
	} else
	    *out++ = *in;
	in++;
	candot = 1;
    }
    *out = '\0';
    return;
}

static void unmungestr(char *in, char *out, int outlen)
{
    while (*in) {
	if (*in == '%' && in[1] && in[2]) {
	    int i, j;

	    i = in[1] - '0';
	    i -= (i > 9 ? 7 : 0);
	    j = in[2] - '0';
	    j -= (j > 9 ? 7 : 0);

	    *out++ = (char) ((i << 4) + j);
	    if (!--outlen)
		return;
	    in += 3;
	} else {
	    *out++ = *in++;
	    if (!--outlen)
		return;
	}
    }
    *out = '\0';
    return;
}
#endif

void *open_settings_w(char * /*sessionname*/)
{
    return (void *) NULL;
}

void write_setting_s(void * /*handle*/, char * /*key*/, char * /*value*/)
{
}

void write_setting_i(void * /*handle*/, char * /*key*/, int /*value*/)
{
}

void close_settings_w(void * /*handle*/)
{
}

void *open_settings_r(char * /*sessionname*/)
{
    return (void *) NULL;
}

char *read_setting_s(void * /*handle*/, char * /*key*/, char * /*buffer*/, int /*buflen*/)
{
	return NULL;
}

int read_setting_i(void * /*handle*/, char * /*key*/, int defvalue)
{
	return defvalue;
}

void close_settings_r(void * /*handle*/)
{
}

void del_settings(char * /*sessionname*/)
{
}

struct enumsettings {
    int i;
};

void *enum_settings_start(void)
{
	return NULL;
}

char *enum_settings_next(void * /*handle*/, char * /*buffer*/, int /*buflen*/)
{
	return NULL;

}

void enum_settings_finish(void * /*handle*/)
{
}

/*
 * Returns:
 * 2,	if key is different in db
 * 1,	if key does not exist in db
 * 0, 	if key matched OK in db
 */

int VerifyHostKeyL(char *hostname, int port, char *keytype, char *key) {

    // Construct host key file name and check if it exists.
    TFindFile ff(fs);
    assert(dataPath.Length() > 0);
    if ( ff.FindByDir(KHostKeysFile, dataPath) != KErrNone ) {
        return 1; // not in database
    }

    // Open database file
    CLineReader *reader = CLineReader::NewLC(ff.File(), fs);
    
    // Construct key ID and key descriptors
    HBufC8 *keyId = HBufC8::NewLC(strlen(hostname) + strlen(keytype) + 12);
    _LIT8(KKeyFormat, "%s:%d:%s");
    keyId->Des().Format(KKeyFormat, hostname, port, keytype);
    TPtrC8 keyDes((TUint8*) key, strlen(key));

    // Go through the file, looking for this key
    TInt result = -1;
    while ( result == -1 ) {
        TPtrC8 &line = reader->ReadLineL();
        if ( line.Length() == 0 ) {
            result = 1; // not in database
        } else {
            // Check if the line starts with the key ID
            int firstSpace = line.Locate(' ');
            if ( firstSpace == KErrNotFound ) {
                continue;
            }
            if ( line.Left(firstSpace).Compare(*keyId) == 0 ) {
                // ID matches. Check if they key matches too
                TInt keyLen = line.Length() - firstSpace - 1;
                if ( (keyLen > 0 ) &&
                     line.Right(keyLen).Compare(keyDes) == 0 ) {
                    result = 0; // key matched OK
                } else {
                    result = 2; // key is different in database
                }
            }
        }
    }

    CleanupStack::PopAndDestroy(2); // keyid, reader
    return result;
}


int verify_host_key(char *hostname, int port, char *keytype, char *key) {

    TInt res = 0;
    TRAPD(error, res = VerifyHostKeyL(hostname, port, keytype, key));
    if ( error != KErrNone ) {
        fatalbox("verify_host_key: Error %d", error);
    }
    return res;
}


void StoreHostKeyL(char *hostname, int port, char *keytype, char *key) {

    _LIT8(KCRLF, "\r\n");    

    // Check if the file exists. If not, create it. Determine full file name.
    TBool haveHostFile = EFalse;
    TFindFile ff(fs);
    TFileName fileName;
    assert(dataPath.Length() > 0);
    if ( ff.FindByDir(KHostKeysFile, dataPath) != KErrNone ) {
        fs.MkDirAll(dataPath);
        fileName = dataPath;
        fileName.Append(KHostKeysFile);
    } else {
        fileName = ff.File();
        haveHostFile = ETrue;
    }

    // Build the new key database entry and key ID
    HBufC8 *newKey = HBufC8::NewLC(strlen(hostname)+strlen(keytype)+strlen(key)+14);
    _LIT8(KKeyFormat, "%s:%d:%s %s");
    newKey->Des().Format(KKeyFormat, hostname, port, keytype, key);
    
    HBufC8 *keyId = HBufC8::NewLC(strlen(hostname) + strlen(keytype) + 12);
    _LIT8(KIdFormat, "%s:%d:%s");
    keyId->Des().Format(KIdFormat, hostname, port, keytype);

    // We'll go through the key database file, copy all other keys to a
    // temporary file, and replace the key for this host/port/type with a new
    // one. If it doesn't exist, we'll append it to the end.
    RFile tempFile;
    TFileName tempFileName;
    tempFileName = dataPath;
    tempFileName.Append(KTmpHostKeysFile);
    User::LeaveIfError(tempFile.Replace(fs, tempFileName,
                                        EFileWrite | EFileShareExclusive));
    CleanupClosePushL(tempFile);
    TBool keyWritten = EFalse;

    if ( haveHostFile ) {
        
        CLineReader *reader = CLineReader::NewLC(fileName, fs);
    
        TBool atEof = EFalse;
        while ( !atEof ) {
            TPtrC8 &line = reader->ReadLineL();
            if ( line.Length() == 0 ) {
                // EOF
                atEof = ETrue;
                break;
            }
        
            // Check if the line starts with the key ID
            int firstSpace = line.Locate(' ');
            if ( (firstSpace != KErrNotFound) &&
                 (line.Left(firstSpace).Compare(*keyId) == 0) ) {
            
                // Yes, we'll replace this line with the new key
                User::LeaveIfError(tempFile.Write(*newKey));
                User::LeaveIfError(tempFile.Write(KCRLF));
                keyWritten = ETrue;
            
            } else {
                // No, just write this line back
                User::LeaveIfError(tempFile.Write(line));
                User::LeaveIfError(tempFile.Write(KCRLF));
            }
        }

        CleanupStack::PopAndDestroy(); // reader        
    }

    // If we didn't manage to write the key yet, do it now
    if ( !keyWritten ) {
        User::LeaveIfError(tempFile.Write(*newKey));
        User::LeaveIfError(tempFile.Write(KCRLF));
    }

    CleanupStack::PopAndDestroy(); // tempFile;

    // Replace the old key file with the new one
    User::LeaveIfError(fs.Replace(tempFileName, fileName));

    // Done
    CleanupStack::PopAndDestroy(2); // keyId, newkey
}

void store_host_key(char *hostname, int port, char *keytype, char *key) {

    TRAPD(error, StoreHostKeyL(hostname, port, keytype, key));
    if ( error != KErrNone ) {
        fatalbox("store_host_key: Error %d", error);
    }
}

void read_random_seed(noise_consumer_t consumer)
{
	TInt ret;
	TFindFile ff(fs);
	RFile f;
	TBuf8<200> buf;

        assert(dataPath.Length() > 0);
	ret=ff.FindByDir(KRandomFile, dataPath);
	if (ret==KErrNotFound)
	{
		return;
	}
	User::LeaveIfError(f.Open(fs, ff.File(), EFileRead));
	while(f.Read(buf)==KErrNone)
	{
		if (buf.Length()==0) break;
		consumer((void *)buf.Ptr(), buf.Length());
	}
	f.Close();
}

void write_random_seed(void *data, int len)
{
	TInt ret;
	TFindFile ff(fs);
	TParse fp;
	RFile f;
	TPtrC8 d((unsigned char *)data,len);

        assert(dataPath.Length() > 0);
	ret=ff.FindByDir(KRandomFile,dataPath);
	if (ret==KErrNotFound)
	{
		fs.MkDirAll(dataPath);
		User::LeaveIfError(fp.Set(KRandomFile,&dataPath,NULL));
	}
	else User::LeaveIfError(fp.Set(ff.File(),NULL,NULL));
	User::LeaveIfError(f.Replace(fs, fp.FullName(), EFileWrite));
	User::LeaveIfError(f.Write(d));
	User::LeaveIfError(f.Flush());
	f.Close();
}

void cleanup_all(void)
{
}
